#!/usr/bin/env python
"""
# Author: Pan Jia, Wenjun Liu, Ning Liu, and Fuyi Li
# File Name: __init__.py
# Description:
"""

__author__ = "Pan Jia, Wenjun Liu, Ning Liu, and Fuyi Li"
__email__ = "jiapan@nwafu.edu.cn; nora.liu@adelaide.edu.au; ning.liu@adelaide.edu.au; fuyi.li@nwafu.edu.cn"


from .cluster import clustering
from .preprocess import optimized_construct_interaction, preprocess, grid_downsample, get_feature, add_contrastive_label, preprocess_adj_sparse


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.module import Module
    
from domino_spatial.layer import AvgReadout, Discriminator

class GCNEncoder(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim, dropout=0.5, act=F.relu):
        super(GCNEncoder, self).__init__()
        self.linear = nn.Linear(input_dim, output_dim)
        self.dropout = dropout
        self.act = act

    def forward(self, x, adj_sparse):
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.linear(x)
        # Sparse multiplication
        if x.is_cuda:
            with torch.cuda.amp.autocast(enabled=False):
                x = torch.sparse.mm(adj_sparse.float(), x.float())
            x = x.to(x.dtype)  # cast back (bf16/fp16-safe)
        else:
            x = torch.sparse.mm(adj_sparse, x)
        return self.act(x)

class SharedMLP(nn.Module):
    '''A shared MLP with two hidden layers and PReLU activation.'''
    def __init__(self, input_dim, hidden_dim, proj_dim):
        super(SharedMLP, self).__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, proj_dim)
        self.prelu = nn.PReLU()
        
    def forward(self, x):
        orig_dtype = x.dtype
        x = self.fc1(x)

        if x.is_cuda:
            # PReLU has no bf16/fp16 CUDA kernel: disable autocast and upcast to fp32
            with torch.cuda.amp.autocast(enabled=False):
                x32 = torch.nn.functional.prelu(x.float(), self.prelu.weight.float())
            x = x32.to(orig_dtype)  
        else:
            # CPU path (fp32)
            x = torch.nn.functional.prelu(x, self.prelu.weight)
        x = self.fc3(x)
        return x


    
class GDCGraphCL(Module):
    '''
    A multi-view graph contrastive learning framework based on graph diffusion
    
    Parameters
    ----------
    input_dim : 
        Input feature dimension.
    hidden_dim : 
        Hidden layer dimension.
    output_dim : 
        Output embedding dimension.
    proj_dim : 
        Projection head dimension.
    graph_neigh : 
        Nearest neighbor adjacency matrix.
    graph_diffusion : 
        Diffusion-processed adjacency matrix.
    dropout : 
        Dropout rate.
    act : 
        Activation function.

    Returns
    -------
    emb: 
        Reconstructed feature.
    ret: 
        Discriminator consistency score between the original view node representation and the graph diffusion view graph representation.
    ret_a: 
        Discriminator consistency score between the graph diffusion view node representation and the original view graph representation.
    
    '''
    def __init__(self, input_dim, hidden_dim, output_dim, proj_dim, graph_neigh, graph_diffusion, dropout=0.5, act=F.relu):
        super(GDCGraphCL, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.proj_dim = proj_dim
        self.graph_neigh = graph_neigh
        self.graph_diffusion = graph_diffusion
        self.dropout = dropout
        self.act = act
        
        self.encoder = GCNEncoder(input_dim, hidden_dim, output_dim, dropout, act) 
        self.decoder = GCNEncoder(output_dim, hidden_dim, input_dim, dropout, act)
        self.mlp = SharedMLP(output_dim, hidden_dim, proj_dim)  

        self.disc = Discriminator(proj_dim)
        self.sigm = nn.Sigmoid()
        self.read = AvgReadout()
    
    # when reading sparse graphs, using this to avoid dense
    def _sparse_avg(self, H, A_sp):
        deg = torch.sparse.sum(A_sp, dim=1).to_dense().clamp_min(1.0)
        if H.is_cuda:
            with torch.cuda.amp.autocast(enabled=False):
                out = torch.sparse.mm(A_sp.float(), H.float())
            out = out.to(H.dtype)
            deg = deg.to(H.dtype)
        else:
            out = torch.sparse.mm(A_sp, H)
        return out / deg.unsqueeze(1)   

    def encode(self, feat, adj):
        return self.encoder(feat, adj)

    def decode(self, h, adj):
        return self.decoder(h, adj)

    def forward(self, feat, adj, gdc_adj):
        # Original view
        h = self.encode(feat, adj)
        # Diffusion-augmented view
        h_g = self.encode(feat, gdc_adj) 
        # Feature-shuffled negative sample
        perm = torch.randperm(feat.size(0), device=feat.device)
        shuf_h = self.encode(feat[perm], adj)
        # Reconstructed feature
        emb = self.decode(h, adj)

        z = self.mlp(h)
        z_g = self.mlp(h_g)
        shuf_z = self.mlp(shuf_h)
        
        # Original view graph representation
        g = self._sparse_avg(h, self.graph_neigh) 
        g = self.sigm(g)
        g = self.mlp(g)  

        # Diffusion-augmented view graph representation
        g_g = self._sparse_avg(h_g, self.graph_diffusion)
        g_g = self.sigm(g_g)  
        g_g = self.mlp(g_g)

        ret = self.disc(g_g, z, shuf_z)  
        ret_a = self.disc(g, z_g, shuf_z) 
        
        return emb, ret, ret_a
    





# Implementation of functions: get_feature, add_contrastive_label and preprocess_adj_sparse is referencing from GraphST: https://doi.org/10.1038/s41467-023-36796-3

import numpy as np
import pandas as pd
import scanpy as sc
import torch
import scipy.sparse as sp
from scipy.sparse import csc_matrix
from scipy.sparse import csr_matrix
from sklearn.neighbors import NearestNeighbors 
from scipy import sparse
from scipy.sparse import eye, diags
from scipy.sparse import coo_matrix

def _arnoldi_iteration(T, alpha, max_iter, tol, eps=1e-6):
    """Arnoldi iteration method for accelerating convergence."""
    N = T.shape[0]
    S = alpha * sp.eye(N, dtype=np.float32, format='csr')
    X = (1 - alpha) * T
    
    for _ in range(max_iter):
        delta = alpha * X
        S += delta

        # Immediately filter out the tiny values after each iteration.
        delta.data[delta.data < eps] = 0
        delta.eliminate_zeros()
        
        # Check for convergence.
        if delta.nnz == 0 or np.abs(delta).sum() < tol * N:
            break
            
        # Optimize using matrix-vector product.
        X = T @ X  
        
    return S

def optimized_balanced_gdc(A, alpha=0.15, eps=1e-6, n_iter=35, tol=1e-5, k=50):
    """optimized balance of PPR diffusion."""
    N = A.shape[0]
    
    # Convert to CSR format to achieve the best matrix multiplication performance.
    if not isinstance(A, csr_matrix):
        A = csr_matrix(A)
    
    # Add self-loops and normalize.
    A_loop = A + eye(N, dtype=np.float32, format='csr')
    D_loop_vec = A_loop.sum(axis=1).A.ravel().astype(np.float32)
    D_loop_invsqrt = diags(1 / np.sqrt(D_loop_vec + 1e-12), dtype=np.float32, format='csr')
    # T_sym = D_loop_invsqrt @ A_loop @ D_loop_invsqrt
    T_sym = D_loop_invsqrt.dot(A_loop).dot(D_loop_invsqrt)

    # Perform Arnoldi iteration.
    S = _arnoldi_iteration(T_sym, alpha, n_iter, tol)
    
    # Efficient sparsification and normalization.
    S.data[S.data < eps] = 0
    S.eliminate_zeros()

    # Enforce sparsity by keeping top-k edges per node.
    if k is not None:
        rows, cols = S.nonzero()
        data = S.data
        coo_dict = {}
        for i, j, v in zip(rows, cols, data):
            if i not in coo_dict:
                coo_dict[i] = []
            coo_dict[i].append((j, v))
        
        # Keep top-k edges per node.
        new_rows, new_cols, new_data = [], [], []
        for i in coo_dict:
            edges = sorted(coo_dict[i], key=lambda x: -x[1])[:k]
            for j, v in edges:
                new_rows.append(i)
                new_cols.append(j)
                new_data.append(v)
        
        S = sp.coo_matrix((new_data, (new_rows, new_cols)), shape=(N, N)).tocsr()
    
    # Use a more stable normalization method.
    D_tilde_vec = S.sum(axis=1).A.ravel()
    D_tilde_inv = diags(1 / np.maximum(D_tilde_vec, 1e-12), dtype=np.float32, format='csr')
    # return S @ D_tilde_inv
    return S.dot(D_tilde_inv)

def build_sparse_adjacency(position, n_neighbors=5):
    """Efficient construction of sparse adjacency matrix."""
    n_spot = position.shape[0]
    
    # Using BallTree to find neighbors.
    nn = NearestNeighbors(n_neighbors=n_neighbors+1, algorithm='ball_tree')
    nn.fit(position)
    distances, indices = nn.kneighbors(position)
    
    # Construct the original neighbor matrix (asymmetric).
    rows = np.repeat(np.arange(n_spot), n_neighbors)
    cols = indices[:, 1:].flatten()  
    data = np.ones_like(rows, dtype=np.int8)
    
    # Create the original neighbor matrix (in COO format).
    adj_original = coo_matrix((data, (rows, cols)), shape=(n_spot, n_spot))
    adj_original = adj_original.tocsr()  
    
    # Obtain the symmetric adjacency matrix by adding the transpose.
    adj_sym = adj_original + adj_original.T
    adj_sym.data = np.ones_like(adj_sym.data)  
    
    return adj_sym, adj_original

def optimized_construct_interaction(adata, n_neighbors=5, alpha=0.15, eps=1e-6, n_iter='auto', tol=1e-5):
    """Optimized function for constructing spatial interaction graphs with diffusion processing."""
    position = adata.obsm['spatial']
    n_spot = position.shape[0]
    
    # Automatically determine iteration count based on number of spots.
    if n_iter == 'auto':
        n_iter = 40 if n_spot < 2000 else 30 if n_spot < 5000 else 25
    
    # Build sparse adjacency matrix from spatial coordinates.
    adj, interaction = build_sparse_adjacency(position, n_neighbors=n_neighbors)

    # Original neighbor matrix
    adata.obsm['graph_neigh'] = interaction
    # Symmetric adjacency matrix
    adata.obsm['adj'] = adj

    # Graph diffusion processing.
    graph_diffusion = optimized_balanced_gdc(
        adj, 
        alpha=alpha,
        eps=eps,
        n_iter=n_iter,
        tol=tol
    )
    # Neighbor matrix after diffusion processing.
    adata.obsm['graph_diffusion'] = graph_diffusion

    # Symmetric adjacency matrix after diffusion processing.
    adj_diffusion = 0.5 * (graph_diffusion + graph_diffusion.T)
    adj_diffusion.data = (adj_diffusion.data > 0).astype(np.int8)
    adata.obsm['adj_diffusion'] = adj_diffusion
    
    return adata

def preprocess(adata):
    sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=2000)
    adata = adata[:, adata.var['highly_variable']].copy()
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.scale(adata, zero_center=False, max_value=10)

    return adata

def grid_downsample(
    adata, 
    grid_size=(100, 100), 
    downsample_by='median',  
    keep_sparse=True         
):
    """
    Parameters
    ----------
    adata : 
        AnnData object of spatial data.
    grid_size : 
        Grid division size (x_bins, y_bins).
    downsample_by : 
        Sampling method ('random'/'median').
    keep_sparse : 
        Maintain the sparse matrix format or not.
        
    Returns
    -------
    AnnData object after downsampling.
    
    """
    coords = adata.obsm['spatial']
    min_x, min_y = coords.min(axis=0)
    max_x, max_y = coords.max(axis=0)
    
    # Generate grid boundaries.
    eps_x, eps_y = (max_x-min_x)*1e-5, (max_y-min_y)*1e-5
    x_bins = np.linspace(min_x, max_x+eps_x, grid_size[0]+1)
    y_bins = np.linspace(min_y, max_y+eps_y, grid_size[1]+1)
    
    # Calculate grid coordinates (vectorization).
    coords = np.asarray(coords)
    x_idx = np.digitize(coords[:,0], x_bins) - 1
    y_idx = np.digitize(coords[:,1], y_bins) - 1
    grid_ids = x_idx * grid_size[1] + y_idx
    
    # Group processing.
    unique_ids = pd.unique(grid_ids)
    sampled_indices = []
    grid_assignments = np.full(len(coords), -1, dtype=int)
    
    for i, cell_id in enumerate(unique_ids):
        cell_mask = (grid_ids == cell_id)
        cell_indices = np.where(cell_mask)[0]
        
        if len(cell_indices) == 0:
            continue
            
        # Select representative points based on the sampling strategy.
        if downsample_by == 'random':
            selected_idx = np.random.choice(cell_indices)
        elif downsample_by == 'median':
            # Calculate the median coordinates of all points within the grid.
            median_coord = np.median(coords[cell_indices], axis=0)
            # Select the point that is closest to the median distance.
            distances = np.linalg.norm(coords[cell_indices] - median_coord, axis=1)
            selected_idx = cell_indices[np.argmin(distances)]
        else:
            raise ValueError("downsample_by must be 'random' or 'median'")
            
        sampled_indices.append(selected_idx)
        grid_assignments[cell_mask] = i
    
    # Constructing downsampled data.
    adata_downsampled = adata[sampled_indices].copy()
    
    # Processing matrix format.
    if keep_sparse and sparse.issparse(adata.X):
        adata_downsampled.X = adata.X[sampled_indices].tocsr()
    
    # Preserve the original spatial coordinates and grid assignments.
    adata_downsampled.obs['grid_id'] = grid_ids[sampled_indices]
    adata_downsampled.uns['grid_assignments'] = grid_assignments
    adata_downsampled.obsm['spatial'] = coords[sampled_indices]
    
    return adata_downsampled

def get_feature(adata):
    '''Extracts and processes feature matrix from AnnData object.'''
    adata_Vars = adata
       
    if isinstance(adata_Vars.X, csc_matrix) or isinstance(adata_Vars.X, csr_matrix):
       feat = adata_Vars.X.toarray()[:, ]
    else:
       feat = adata_Vars.X[:, ] 
    
    adata.obsm['feat'] = feat
    
def add_contrastive_label(adata):
    '''Adds contrastive label to AnnData object.'''
    n_spot = adata.n_obs
    one_matrix = np.ones([n_spot, 1])
    zero_matrix = np.zeros([n_spot, 1])
    label_CSL = np.concatenate([one_matrix, zero_matrix], axis=1)
    adata.obsm['label_CSL'] = label_CSL
    
def preprocess_adj_sparse(adj):
    """Convert scipy sparse matrix to torch.sparse.FloatTensor."""
    adj = sp.coo_matrix(adj)
    adj_ = adj + sp.eye(adj.shape[0])
    rowsum = np.array(adj_.sum(1))
    degree_mat_inv_sqrt = sp.diags(np.power(rowsum, -0.5).flatten())
    adj_normalized = adj_.dot(degree_mat_inv_sqrt).transpose().dot(degree_mat_inv_sqrt).tocoo()
    
    indices = torch.from_numpy(np.vstack([adj_normalized.row, adj_normalized.col])).long()
    values = torch.from_numpy(adj_normalized.data).float()
    shape = torch.Size(adj_normalized.shape)

    return torch.sparse.FloatTensor(indices, values, shape)




import argparse
from . import __version__

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    args = parser.parse_args()

if __name__ == "__main__":
    main()


import torch
from domino_spatial.model import GDCGraphCL
from tqdm import tqdm
from torch import nn
import contextlib

from domino_spatial.preprocess import preprocess, preprocess_adj_sparse, get_feature, add_contrastive_label, grid_downsample, optimized_construct_interaction

import warnings
warnings.filterwarnings('ignore')   


class DOMINO():
    def __init__(self, 
        adata,
        device,
        learning_rate=0.001,
        weight_decay=1e-5,
        epochs=800,
        hidden_dim=512,
        output_dim=256,
        proj_dim=128,
        a=1,
        b=1,
        is_downsample=False,
        grid_size=100
        ):
        '''

        Parameters
        ----------
        adata : anndata
            AnnData object of spatial data.
        device : string, optional
            Using GPU or CPU? The default is 'cuda'.
        learning_rate : float, optional
            Learning rate for the Adam optimizer. The default is 0.001.
        weight_decay : float, optional
            Weight decay (L2 penalty) for regularization. The default is 1e-5.
        epochs : int, optional
            Number of training epochs. The default is 800.
        hidden_dim : int, optional
            Dimension of hidden layers in the GNN model. The default is 512.
        output_dim : int, optional
            Dimension of the output embedding. The default is 256.
        proj_dim : int, optional
            Dimension of the projection head for the mlp layers. The default is 128.
        a : int, optional
            Weight coefficient for feature reconstruction loss (loss_feat). The default is 1.
        b : int, optional
            Weight coefficient for graph contrastive learning loss (loss_sl_1 + loss_sl_2). The default is 1.
        is_downsample : bool, optional
            Whether to use downsampling or not. The default is False.
        grid_size : int, optional
            Grid size for downsampling. The default is 100.

        Returns
        -------
        adata : anndata
            AnnData object of spatial data with added 'emb' key in obsm.
        '''

        self.adata = adata.copy()
        self.device = device
        self.learning_rate=learning_rate
        self.weight_decay=weight_decay
        self.epochs=epochs
        self.hidden_dim=hidden_dim
        self.output_dim=output_dim
        self.proj_dim=proj_dim
        self.a=a
        self.b=b
        self.is_downsample=is_downsample
        self.grid_size = grid_size

        torch.backends.cuda.matmul.allow_tf32 = True
        try:
            torch.backends.cudnn.allow_tf32 = True
        except Exception:
            pass
        try:
            torch.set_float32_matmul_precision("medium")
        except Exception:
            pass

        # Autocast context: bf16 on CUDA, no-op on CPU
        self.use_amp = (device.type == "cuda" and torch.cuda.is_bf16_supported())
        self.autocast_ctx = torch.cuda.amp.autocast(dtype=torch.bfloat16) if self.use_amp else contextlib.nullcontext()
        
        self.adata = preprocess(self.adata)

        if self.is_downsample:
            self.adata = grid_downsample(self.adata, (self.grid_size, self.grid_size))

        print("Constructing interaction matrix...")
        self.adata = optimized_construct_interaction(self.adata)

        print("Constructing interaction matrix Done!")

        add_contrastive_label(self.adata)
        get_feature(self.adata)

        self.features = torch.FloatTensor(self.adata.obsm['feat'].copy()).to(self.device)
        self.input_dim = self.features.shape[1]
        self.label_CSL = torch.FloatTensor(self.adata.obsm['label_CSL']).to(self.device)
        # Symmetric adjacency for neighborhood aggregation
        self.adj = self.adata.obsm['adj']
        self.adj_diffusion = self.adata.obsm['adj_diffusion']
        
        # Adjacency matrix, used as a mask for pooling operations
        print("generating graphs..")

        self.graph_neigh = self._add_self_loops_sparse(self._to_sparse_coo(self.adata.obsm['graph_neigh'])).to(self.device)
        self.graph_diff  = self._add_self_loops_sparse(self._to_sparse_coo(self.adata.obsm['graph_diffusion'])).to(self.device)

        self.adj = preprocess_adj_sparse(self.adj).to(self.device)

        self.adj_diffusion = preprocess_adj_sparse(self.adj_diffusion).to(self.device)

    # function for stop making dense matrix when generating graphs
    def _to_sparse_coo(self, A):
        import scipy.sparse as sp
        if hasattr(A, "tocoo"):             # scipy.sparse
            A = A.tocoo()
            import numpy as np
            idx = torch.from_numpy(np.vstack([A.row, A.col]).astype("int64"))
            val = torch.from_numpy(A.data.astype("float32"))
            return torch.sparse_coo_tensor(idx, val, A.shape).coalesce()
        elif isinstance(A, np.ndarray):
            # if dense，then make it sparse
            ii, jj = np.nonzero(A)
            val = A[ii, jj].astype("float32")
            idx = torch.from_numpy(np.vstack([ii, jj]).astype("int64"))
            val = torch.from_numpy(val)
            return torch.sparse_coo_tensor(idx, val, A.shape).coalesce()
        elif isinstance(A, torch.Tensor) and A.is_sparse:
            return A.coalesce()
        else:
            raise TypeError(f"Unsupported adjacency type: {type(A)}")

    def _add_self_loops_sparse(self, A_sp):
        n = A_sp.shape[0]
        idx = torch.arange(n, device=A_sp.device)
        I = torch.sparse_coo_tensor(torch.stack([idx, idx]), torch.ones(n, device=A_sp.device, dtype=A_sp.dtype), (n, n))
        return (A_sp + I).coalesce()

    def mse_chunked(self, target, pred, chunk=200_000):
        """
        Mean((target - pred)^2) without materializing the full N×F diff.
        Exactly matches F.mse_loss(..., reduction='mean').
        Accumulates in float32 for stability (esp. under AMP/bfloat16).
        """
        assert target.shape == pred.shape
        n = target.shape[0]
        total = target.numel()

        # accumulate in fp32, regardless of AMP autocast dtype
        acc = pred.new_zeros((), dtype=torch.float32)

        for start in range(0, n, chunk):
            stop = min(start + chunk, n)
            diff = (pred[start:stop] - target[start:stop]).to(torch.float32)
            acc = acc + (diff * diff).sum()

        return acc / total 

    def train(self):
        # Initialize the model 
        print("Initializing model..")
        self.model = GDCGraphCL(self.input_dim, self.hidden_dim, self.output_dim, self.proj_dim, self.graph_neigh, self.graph_diff).to(self.device)

        # Binary cross-entropy loss for contrastive learning
        self.loss_CSL = nn.BCEWithLogitsLoss()
        self.optimizer = torch.optim.Adam(self.model.parameters(), self.learning_rate, weight_decay = self.weight_decay)

        print('Begin to train ST data...')
        self.model.train()
        
        for epoch in tqdm(range(self.epochs)): 
            self.model.train()
            self.optimizer.zero_grad(set_to_none=True)    
            
            with self.autocast_ctx:
                self.emb, ret, ret_a = self.model(self.features, self.adj, self.adj_diffusion)
                # Calculate loss 
                self.loss_sl_1 = self.loss_CSL(ret, self.label_CSL)  # Graph contrastive loss for original view
                self.loss_sl_2 = self.loss_CSL(ret_a, self.label_CSL)  # Graph contrastive loss for augmented view
                self.loss_feat = self.mse_chunked(self.features, self.emb)  # Feature reconstruction loss
                # Total loss
                loss =  self.a*self.loss_feat + self.b*(self.loss_sl_1 + self.loss_sl_2)

            if epoch % 100 == 0:
                print(
                    'Epoch {:0>3d} | Loss:[{:.4f}], loss_feat:[{:.4f}], loss_sl_1:[{:.4f}], loss_sl_2:[{:.4f}]'.format(
                        epoch, loss.item(), self.loss_feat.item(), self.loss_sl_1.item(), self.loss_sl_2.item()))
            
            loss.backward() 
            self.optimizer.step()
        
        print("Optimization finished for ST data!")
        
        with torch.no_grad():
            self.model.eval()
            self.emb_rec = self.model(self.features, self.adj, self.adj_diffusion)[0].detach().cpu().numpy()
            self.adata.obsm['emb'] = self.emb_rec
                
            return self.adata



# Implementation of class: AvgReadout and Discriminator is referencing from GraphST: https://doi.org/10.1038/s41467-023-36796-3

import torch
import torch.nn as nn
import torch.nn.functional as F

class AvgReadout(nn.Module):
    def __init__(self):
        super(AvgReadout, self).__init__()

    def forward(self, emb, mask=None):
        # Masked weighted summation
        vsum = torch.mm(mask, emb)
        row_sum = torch.sum(mask, 1)
        # Adjust the shape
        row_sum = row_sum.expand((vsum.shape[1], row_sum.shape[0])).T  
        # Weighted average
        global_emb = vsum / row_sum
          
        # L2 Normalization
        return F.normalize(global_emb, p=2, dim=1)  
    
class Discriminator(nn.Module):
    def __init__(self, n_h):
        super(Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c, h_pl, h_mi, s_bias1=None, s_bias2=None):
        c_x = c.expand_as(h_pl)  

        # Positive score
        sc_1 = self.f_k(h_pl, c_x)
        # Negative score
        sc_2 = self.f_k(h_mi, c_x)

        if s_bias1 is not None:
            sc_1 += s_bias1
        if s_bias2 is not None:
            sc_2 += s_bias2

        # Concatenate positive and negative scores
        logits = torch.cat((sc_1, sc_2), 1)

        return logits



# Implementation of functions: mclust_R and search_res is referencing from GraphST: https://doi.org/10.1038/s41467-023-36796-3

import numpy as np
import pandas as pd
import scanpy as sc
import ot
from sklearn.decomposition import PCA
from sklearn.cluster import MiniBatchKMeans
import ot.backend as otb

def mclust_R(adata, num_cluster, modelNames='EEE', used_obsm='emb_pca', random_seed=2020):
    """
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """
    
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']
    
    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    return adata

def clustering(adata, radius=50, n_clusters=7, method='mclust', start=0.1, end=3.0, increment=0.01, refinement=False):
    """
    Spatial clustering based the learned representation.    

    Parameters
    ----------
    adata : 
        AnnData object of scanpy package.
    n_clusters : 
        The number of clusters. The default is 7.
    radius : 
        The number of neighbors considered during refinement. The default is 50.
    key : 
        The key of the learned representation in adata.obsm. The default is 'emb'.
    method : 
        The tool for clustering. Supported tools include 'mclust', 'leiden', and 'louvain'. The default is 'mclust'. 
    start : 
        The start value for searching. The default is 0.1.
    end : 
        The end value for searching. The default is 3.0.
    increment : 
        The step size to increase. The default is 0.01.   
    refinement : 
        Refine the predicted labels or not. The default is False.

    Returns
    -------
    None.

    """
    pca = PCA(n_components=20, random_state=42) 
    embedding = pca.fit_transform(adata.obsm['emb'].copy())
    adata.obsm['emb_pca'] = embedding
  
    if method == 'mclust':
       adata = mclust_R(adata, num_cluster = n_clusters, modelNames='EEE', used_obsm='emb_pca', random_seed=2020)
       adata.obs['domain'] = adata.obs['mclust']
    elif method == 'leiden':
       res = search_res(adata, n_clusters, use_rep='emb_pca', method=method, start=start, end=end, increment=increment)
       sc.tl.leiden(adata, random_state=0, resolution=res)
       adata.obs['domain'] = adata.obs['leiden']
    elif method == 'louvain':
       res = search_res(adata, n_clusters, use_rep='emb_pca', method=method, start=start, end=end, increment=increment)
       sc.tl.louvain(adata, random_state=0, resolution=res)
       adata.obs['domain'] = adata.obs['louvain'] 
    elif method == 'kmeans':
        # MiniBatchKMeans suitable for large datasets
        kmeans = MiniBatchKMeans(
            n_clusters=n_clusters,
            batch_size=1024,
            n_init=10,
            init='k-means++',
            max_iter=300,
            tol=1e-3,
            random_state=42
        )
        adata.obs['domain'] = kmeans.fit_predict(adata.obsm['emb_pca']).astype(str)

    if refinement:  
       new_type = refine_label(adata, radius, key='domain')
       adata.obs['domain'] = new_type
        
# using blocking to reduce memory usage
def refine_label(adata, radius=50, key='label', row_block=4096):
    '''Based on the spatial position information of cells, local smoothing and optimization are performed on the initial clustering results.'''
    n_neigh = radius
    new_type = []
    old_type = adata.obs[key].values
    
    print("refining domain..")

    #calculate distance
    position = adata.obsm['spatial']
    if hasattr(position, "toarray"):          
        position = position.toarray()
    elif hasattr(position, "to_numpy"):       
        position = position.to_numpy()
    elif "torch" in type(position).__module__:  
        position = position.detach().cpu().numpy()

    position = np.asarray(position, dtype=np.float64)
    if position.ndim != 2 or position.shape[1] < 2:
        raise ValueError(f"spatial must be (n,>=2), got {position.shape}")
    position = np.ascontiguousarray(position[:, :2]) 

    print("running optimal transport...")
    # use blockwise computation in optimal transport
    # calculating distances only for a subset of rows against 
    # the whole matrix at a time, and keep only the
    # top-k neighbor indices per row, without storing 
    # the full distance matrix in memory.   
    n_cell = position.shape[0]
    if n_cell <= 1 or n_neigh < 1:
        return [str(x) for x in old_type]
    n_neigh = min(n_neigh, n_cell - 1)
    
    cats = pd.Categorical(old_type)
    codes = cats.codes   
    n_classes = len(cats.categories)

    for start in range(0, n_cell, row_block):
        stop = min(start + row_block, n_cell)
        block = position[start:stop]

        D = ot.dist(block, position, metric='euclidean')

        part = np.argpartition(D, kth=n_neigh, axis=1)[:, :n_neigh+1]
        rows = np.arange(part.shape[0])[:, None]
        part = part[rows, np.argsort(D[rows, part], axis=1)]

        for i in range(stop - start):
            idx = part[i]
            self_col = start + i

            if idx[0] == self_col:
                idx = idx[1:n_neigh+1]
            else:
                idx = idx[:n_neigh]
                idx = idx[idx != self_col]

            neigh_codes = codes[idx]
            neigh_codes = neigh_codes[neigh_codes >= 0]
            if neigh_codes.size == 0:
                new_label_code = codes[start + i]
            else:
                cnt = np.bincount(neigh_codes, minlength=n_classes)
                winners = np.flatnonzero(cnt == cnt.max())
                new_label_code = (codes[start + i]
                                   if codes[start + i] in winners
                                   else winners[0])

            new_type.append(cats.categories[new_label_code] if new_label_code >= 0
                            else str(old_type[start + i]))
        del D, part
        
    new_type = [str(i) for i in list(new_type)]

    print("finished refining...")
    return new_type

def search_res(adata, n_clusters, method='leiden', use_rep='emb', start=0.1, end=3.0, increment=0.01):
    '''
    Searching corresponding resolution according to given cluster number
    
    Parameters
    ----------
    adata : 
        AnnData object of spatial data.
    n_clusters : 
        Targetting number of clusters.
    method : 
        Tool for clustering. Supported tools include 'leiden' and 'louvain'. The default is 'leiden'.    
    use_rep : 
        The indicated representation for clustering.
    start : 
        The start value for searching.
    end : 
        The end value for searching.
    increment : 
        The step size to increase.
        
    Returns
    -------
    res : 
        Resolution.
        
    '''
    print('Searching resolution...')
    label = 0
    sc.pp.neighbors(adata, n_neighbors=50, use_rep=use_rep)
    for res in sorted(list(np.arange(start, end, increment)), reverse=True):
        if method == 'leiden':
           sc.tl.leiden(adata, random_state=0, resolution=res)
           count_unique = len(pd.DataFrame(adata.obs['leiden']).leiden.unique())
           print('resolution={}, cluster number={}'.format(res, count_unique))
        elif method == 'louvain':
           sc.tl.louvain(adata, random_state=0, resolution=res)
           count_unique = len(pd.DataFrame(adata.obs['louvain']).louvain.unique()) 
           print('resolution={}, cluster number={}'.format(res, count_unique))
        if count_unique == n_clusters:
            label = 1
            break

    assert label==1, "Resolution is not found. Please try bigger range or smaller step!." 
       
    return res    
       



from setuptools import find_packages, setup

__lib_name__ = "domino-spatial"
__lib_version__ = "0.1.5"
__description__ = "DOMINO: diffusion-optimised graph learning identifies domain structures with enhanced accuracy and scalability"
__url__ = "https://github.com/ABILiLab/DOMINO"
__author__ = "Pan Jia, Wenjun Liu, Ning Liu, and Fuyi Li"
__author_email__ = "jiapan@nwafu.edu.cn; nora.liu@adelaide.edu.au; ning.liu@adelaide.edu.au; fuyi.li@nwafu.edu.cn"
__requires__ = []
__long_description__ = "DOMINO is built based on a self-supervised multi-view graph contrastive learning framework. It is designed to integrate spatial coordinates and gene expression information for robust identification of tissue domains from spatial transcriptomics (ST) data. DOMINO employs graph neural networks (GNNs) as base encoder, constructing a multi-view graph contrastive learning framework using the original graph and the diffusion graph to learn spot representations in the ST data. After representation learning, the learned low-dimensional embeddings can then be used to identify spatial domains, which can further be used in different downstream analyses, including cell type composition analysis, differential expression, and inference of cell–cell communication."

setup(
    name = __lib_name__,
    version = __lib_version__,
    description = __description__,
    url = __url__,
    author = __author__,
    author_email = __author_email__,
    license = "MIT",
    packages = ["domino_spatial"],
    entry_points={
        'console_scripts': [
            'domino-spatial=domino_spatial.cli:main',  
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",          
    install_requires = __requires__,
    zip_safe = False,
    include_package_data = True,
    long_description = __long_description__,
    long_description_content_type="text/markdown"
)
